import random
from structures import Record
from structures import Cluster
from structures import RecordChunk
#Takes the disassociated dataset as input and reconstruct a random dataset that has a
#similar statistical properties close to the original dataset

class DatasetReconstruct:
    

    def reconstructDataset(self, disassociatedDataset):
        #Shuffle the order of the clusters
        random.shuffle(disassociatedDataset)
        records = []
        recNum = 1
        for cluster in disassociatedDataset:
            #Construct an initial list of records of the size of the cluster 
            clusterRecords = []
            #if there is no record chunk, create a list of empty records
            if not cluster.getRecordChunk():
                for i in range(len(cluster.getRecords())):
                    #When reconstructing dataset, the recID does not matter
                    clusterRecords.append(Record(recNum))
                    recNum+=1

            else:
                clusterRecords = self.reconstructRecords(cluster)

            self.padTermChunks(cluster, clusterRecords)
            records.extend(clusterRecords)

        #Remove the empty records if exists
        i=0
        for record in records:
            if not record:
                del records[i]
            i+=1
                
        return records


    def reconstructRecords(self, cluster):
        #Shuffle the order of subrecords in each record chunk
        for recordChunk in cluster.getRecordChunk():
            random.shuffle(recordChunk.recordChunkValue)

        iterChunk = iter(cluster.getRecordChunk())
        #Construct the first record chunk as the initial list of records
        firstChunk = next(iterChunk)
        clusterRecords = []
        for record in firstChunk.getRecordChunk():
            clusterRecords.append(record)

        #Then we append the subrecords of the remaining record chunks to the
        #initial list of records
        finloop = False
        while not finloop:
            try:
                nextChunk = next(iterChunk)
            except StopIteration:
                finloop = True

            else:
                for i in range(len(clusterRecords)):
                    clusterRecords[i].extRecValue(nextChunk.getRecordChunk()[i].getRecValue())

        return clusterRecords

    #Pad some term chunks into the list of records
    def padTermChunks(self, cluster, clusterRecords):
        #check if the term chunk is not empty
        if cluster.getTermChunk():
            for record in clusterRecords:
                tempTermChunk = list(cluster.getTermChunk())
                #If the record is empty, then at least one term need to be padded
                if not record.getRecValue():
                    #randomize the number of term to be padded between 1 and the size of term chunk
                    #Set a maximum of 10 terms to prevent overly large and unrealistic records
                    numTerm = random.randint(1, min(10,len(tempTermChunk)))

                else:
                    numTerm = random.randint(0, min(10,len(tempTermChunk)))


                #Randomly choose numTerm of terms from term chunk
                for i in range(numTerm):
                    #randomize an index number
                    index = random.randint(0, len(tempTermChunk)-1)
                    #add the term of that index into record
                    term = tempTermChunk[index]
                    record.addRecValue(term)
                    #remove the term so there won't be duplicate
                    tempTermChunk.remove(term)
        
                    
                    
        
        
        
    
