from structures import Cluster

#Horizontally partitioning dataset into clusters
#Horizon partitioning partitions the dataset into two list of records(one with records
#containing the most frequent term in the dataset, another one is the remaining records) recursively
#If the list contains records less than max cluster size, it becomes a cluster. Otherwise,
#the list will be add to a queue and wait to be partitioned again later

class HorPart:

    def __init__(self, maxClusterSize, k):
        self.k = k
        self.maxClusterSize = maxClusterSize
        self.clustersList = []
        self.clusterCounter = 1
        self.queueListData = []
        self.queueListIgnore = []
        self.rejectedRecords = []

    def horizonPartition(self, dataset: list, ignore: set):
        #get the set of terms in the dataset that are not used in previous partitioning
        remainingTerms = self.getRemainingTerms(dataset, ignore)

        #If the remaining terms are empty, it means that the dataset only contain terms that
        #has previously been used, so we reject these records
        if (len(remainingTerms) == 0):
            for record in dataset:
                self.rejectedRecords.append(dataset)
            return

        #Next we get the most frequent term of the remaining terms list
        mostFreqTerm = self.mostFrequentTerm(remainingTerms)

        #Then we add all records that contains that term to one list, and the rest to another
        freqTermList = []
        otherTermList = []
        for records in dataset:
            if mostFreqTerm in records.getRecValue():
                freqTermList.append(records)

            else:
                otherTermList.append(records)


        #If the list is smaller than k, then they cant form a cluster, and it will be rejected
        #If either of the list is smaller than max cluster size, then we add them as a cluster
        #Otherwise add them to the queue list
        if(len(otherTermList)>0):
            if (len(otherTermList)) < self.k:
                for record in otherTermList:
                    self.rejectedRecords.append(record)
                    
            elif (len(otherTermList)) < self.maxClusterSize:
                self.rejectedRecords = [rec for rec in self.rejectedRecords if rec not in otherTermList]
                self.clustersList.append(Cluster(str(self.clusterCounter), otherTermList))
                self.clusterCounter+=1
            else:
                ignoreCopy = ignore.copy()
                self.queueListData.append(otherTermList)
                self.queueListIgnore.append(ignoreCopy)
        

        if (len(freqTermList)>0):
            if (len(freqTermList)) < self.k:
                for record in freqTermList:
                    self.rejectedRecords.append(record)
                    
            elif (len(freqTermList)) < self.maxClusterSize:
                self.rejectedRecords = [rec for rec in self.rejectedRecords if rec not in freqTermList]
                self.clustersList.append(Cluster(str(self.clusterCounter), freqTermList))
                self.clusterCounter+=1
            else:
                ignoreCopy = ignore.copy()
                ignoreCopy.add(mostFreqTerm)
                self.queueListData.append(freqTermList)
                self.queueListIgnore.append(ignoreCopy)


    def horizonPartitioning(self, dataset: list): #dataset is a list of records
        #if dataset has a size smaller than max cluster size
        if (len(dataset)< self.maxClusterSize):
            self.clustersList.append(Cluster("1",dataset))
            return self.clustersList

        #Put the dataset into the queue list
        ignore = set()
        self.queueListData.append(dataset)
        self.queueListIgnore.append(ignore)

        #Go through the queue list and horizontally partition each of them
        check = True
        while (check):
            #If there are rejected records, we add them into the queue to be partition again later 
            if len(self.rejectedRecords)>0:
                #To avoid indefinite loop in the case of rejectedRecords size is smaller than k
                #We add the records to a cluster that is not full
                if len(self.rejectedRecords)< self.k:
                    for cluster in self.clustersList:
                        if len(cluster.getRecords()) < (self.maxClusterSize - len(self.rejectedRecords)):
                            for record in self.rejectedRecords:
                                cluster.addRecord(record)
                            self.rejectedRecords.clear()
                            break

                    #If there is no cluster that could fit them in, then we add one records at a time
                    #into clusters that are full
                    for cluster in self.clustersList:
                        if len(cluster.getRecords())<(self.maxClusterSize - 1):
                            accepted = []
                            for record in self.rejectedRecords:
                                cluster.addRecord(record)
                                accepted.append(record)
                                if len(cluster.getRecords()) < (self.maxClusterSize - 1):
                                    continue
                                else:
                                    break
                            for a in accepted:
                                self.rejectedRecords.remove(a)

                    self.rejectedRecords.clear()
                
                else:
                    self.queueListData.insert(0, self.rejectedRecords)
                    self.queueListIgnore.insert(0, set())
                    self.rejectedRecords.clear()
            
            while(len(self.queueListData)>0):
                dataset = self.queueListData.pop()
                ignoreT = self.queueListIgnore.pop()
                self.horizonPartition(dataset, ignoreT)

            if (len(self.rejectedRecords)==0):
                break

        return self.clustersList
        
    def getRemainingTerms(self, dataset: list, ignoreTerm: set):
        remTerms = []
        for i in range(len(dataset)):
            terms = dataset[i].getRecValue()
            if isinstance(terms, set):
                for term in terms:
                    remTerms.append(term)
            else:
                remTerms.append(terms)

        rTerm = [r for r in remTerms if not r in ignoreTerm]
        return rTerm

    def mostFrequentTerm(self, remTerms: list):
        count, itm = 0, ''
        dic = {}
        for item in remTerms:
            dic[item] = dic.get(item, 0) + 1
            if dic[item] > count:
                count, itm = dic[item], item
        return(itm)

    def getClustersList(self):
        return self.clustersList

    def getClusterCount(self):
        return self.clusterCounter


    
