#Vertical Partitioning
from itertools import combinations
from structures import Cluster
from structures import Record
from structures import RecordChunk
import collections

class VerPart:
    def __init__(self, k, m):
        self.k = k
        self.m = m

    def vertPartitioning(self, cluster: Cluster):
        termSupportList = self.sortSupportTerm(cluster) #TP
        termChunk = self.createTermChunk(termSupportList) #TT

        #Then we remove the terms that exists in the term chunk from
        #support list(the set of terms of current cluster)
        termSupportList = [term for term in termSupportList if term[0] not in termChunk]       #T remain
        #print (termSupportList)
        termSetList = self.verticalPartition(cluster, termSupportList)
        #print(termSetList)
        recordChunks = self.createRecordChunks(termSetList, cluster)
        cluster.setRecordChunk(recordChunks)
        cluster.setTermChunk(termChunk)
        


    #Sort each term of the cluster with decreasing support
    def sortSupportTerm(self, cluster: Cluster):
        allTerm = []
        records = []
        am = len(cluster.getRecords())
        records = cluster.getRecords()
        for i in range(am):
            terms = records[i].getRecValue()
            if isinstance(terms, set):
                for term in terms:
                    allTerm.append(term)
            else:
                allTerm.append(terms)
                
        termSupportList = collections.Counter(allTerm).most_common()
        return termSupportList


    #Move the terms with support less than k to term chunk
    def createTermChunk(self, supportSet):
        termChunkSet = set({})
        for i in supportSet:
            if i[1] < self.k:
                termChunkSet.add(i[0])

        return termChunkSet


    """For every term in the remaining term set, each term will be added to a test
    chunk, which contains the current chunk and the current term. If the test chunk
    is checked to be k^m anonymous then the current term will be added to the current
    chunk. After adding as many term as possible to the current chunk, while staying
    k^m anonymous, the current chunk become a record chunk, and the terms appeared in
    the current chunk is removed from the remaining term set. Repeat the steps above until
    the remaining term set is empty """
    
    def verticalPartition(self, cluster, tRemain):   
    #tRemain = terms remaining after removing terms that appears in term chunk 
    #           from the set of terms (TP-TT)
    #termSetList = a list of sets of terms in each chunk
        termSetList = []
        #Become a list of terms only (sorted in decreasing support)
        tRemain = [i[0] for i in tRemain]
        while (len(tRemain)>0):
            #Tcurrent (contains the terms that will be assigned to the current set)
            tCurrent = set()
            #A list to record the combinations that are already checked for k^m-anonymity
            oldCombinations = []
            newCombinations = []
            
            for term in tRemain:
                #create a test chunk constructed from Tcurrent union {term}
                chunkTest = tCurrent.copy()
                chunkTest.add(term)
                #If the test chunk only have one term, no need to check, because they have at least k items
                #Only check for combinations
                if (len(chunkTest)>1):
                    testList = list(chunkTest)
                    newCombinations.clear()
                    #Store all the combinations up to length m into the list
                    #Since we only need to know k-anonymity for up to m items
                    for x in range(2, self.m+1):
                        for subset in combinations(testList, x):
                            newCombinations.append(subset)

                    
                    newCombinations = [list(i) for i in newCombinations]

                #If there is only one term in cTest, then automatically add it to Tcurrent, because
                #it is k^m-anonymous (support(term)>k)
                #If cTest is k^m-anonymous then we add the term into Tcurrent

                combinationToCheck = [comb for comb in newCombinations if comb not in oldCombinations]
                if (len(chunkTest) == 1):
                    tCurrent.add(term)

                elif (self.kMAnonymous(term, combinationToCheck, cluster)):
                    for comb in combinationToCheck:
                        oldCombinations.append(comb)
                    tCurrent.add(term)
                    
            #add the set of term in Tcurrent into the list of set of term in the current chunk
            #for later projection
            termSetList.append(tCurrent)
            #remove the terms that appeared in Tcurrent from the remaining terms
            for term in tCurrent:
                tRemain.remove(term)

        return termSetList 
                

    #check if the combincations with the new term are k^m-anonymous 
    def kMAnonymous(self, term, combinationToCheck, cluster):
        #if the combination is not checked before, then check it for k^m-anonymity
        for i in combinationToCheck:
            true = 0            
            recordCount = 0
            records = cluster.getRecords()
            #loop through the records in the cluster
            for record in records:
                recValue = record.getRecValue()
                #if the current record value contains the current combination, then
                #increase the record count
                if all(elem in recValue for elem in i):
                    recordCount+=1
                    #once it reaches k item, we stop and proceed to the next loop
                    if(recordCount == self.k):
                        true = 1
                        break
            #after finishing the for loop, and record count has not reach k
            #means that this term is not k^m-anonymous
            if true == 1:
                continue
            else:
                return False
        
        return True


    def createRecordChunks(self, termSetList, cluster):
        recordChunks = []
        recordChunkID = 1
        if termSetList == None:
            return recordChunks
        for termToProjectSet in termSetList:
            recordChunk = self.projectRecordChunk(termToProjectSet, cluster, recordChunkID)
            recordChunks.append(recordChunk)
            recordChunkID+=1

        return recordChunks

    #project the terms onto a record chunk
    def projectRecordChunk(self, termsToProject, cluster, recordChunkID):
        recordChunk = RecordChunk(recordChunkID, [])
        records = cluster.getRecords()
        for record in records:
            newRecord = Record(record.getRecID())
            recVal = record.getRecValue()
            #If the record contains any term of the record chunk
            #Add the term into the new record value
            for term in recVal:
                if (term in termsToProject):
                    newRecord.addRecValue(term)

            recordChunk.addRecord(newRecord)

        return recordChunk

            
            
            
        
        
        
        
        
        
        
        

    
