import collections
from itertools import combinations
import pandas as pd
from mlxtend.frequent_patterns import apriori
from mlxtend.preprocessing import TransactionEncoder

class Evaluation:

    def averageRelativeError(self, originalDataset, publishedDataset):
        #Sort the domain of the dataset by decreasing support
        termSupport = self.sortSupportTerm(originalDataset)
        #Choose the 200th-220th most frequent terms
        terms = termSupport[200:220]
        terms = [i[0] for i in terms]

        #Create combinations of size two
        combos = []
        for x in range(2, 3):
            for subset in combinations(terms, x):
                combos.append(subset)

        for i in combos:
            combos = [list(i) for i in combos]

        #Calculate relative error for each combination
        reDict = {}     #A dictionary for each combination's relative error value
        for combination in combos:
            rErr = self.relativeError(combination, originalDataset, publishedDataset)
            reDict[tuple(combination)] = rErr

        #Calculate the average value of all the relative errors we got
        if (len(reDict) == 0):
            return 'There is no combination'
        else:
            avgRelativeError = sum(reDict.values()) / float(len(reDict))
        return avgRelativeError
   
        
    #Sort the domain of the dataset by decreasing support
    def sortSupportTerm(self, dataset):
        allTerm = []
        records = []
        am = len(dataset)
        records = dataset
        for i in range(am):
            terms = records[i].getRecValue()
            if isinstance(terms, set):
                for term in terms:
                    allTerm.append(term)
            else:
                allTerm.append(terms)

        termSupport = collections.Counter(allTerm).most_common()
        return termSupport

    #Calculate the relative error of the current combination
    def relativeError(self, combination, originalDataset, publishedDataset):
        #Support of this combination in the original dataset
        ogSupport = 0
        for record in originalDataset:
            recValue = record.getRecValue()
            if all(elem in recValue for elem in combination):
                ogSupport+=1

        #Support of this combination in published/reconstructed dataset
        pSupport = 0
        for record in publishedDataset:
            recValue = record.getRecValue()
            if all(elem in recValue for elem in combination):
                pSupport+=1

        #Relative error calculations
        top = abs(ogSupport - pSupport)
        btm = (ogSupport + pSupport)/2
        re = 0
        if(btm !=0):
            re = top/btm

        return re

    #topKdeviation measures how the top-K frequent itemset of the original dataset
    #change in the published anonymized dataset
    #Here we use an apriori function from mlxtend library to extract frequent dataset
    def topKdeviation(self, k, originalDataset, anonymizedDataset):
        ogData = []
        for record in originalDataset:
            recValue = record.getRecValue()
            ogData.append(recValue)
        #apriori function expects data in an encoded pandas DataFrame
        #tranform the dataset into the right format
        te = TransactionEncoder()
        te1 = te.fit(ogData).transform(ogData, sparse=True)
        df1 = pd.DataFrame.sparse.from_spmatrix(te1, columns=te.columns_)

        anonData = []
        for record in anonymizedDataset:
            recValue = record.getRecValue()
            anonData.append(recValue)
        te2 = te.fit(anonData).transform(anonData, sparse=True)
        df2 = pd.DataFrame.sparse.from_spmatrix(te2, columns=te.columns_)

        #Return the items and itemsets with at least 0.1% support:
        #support = freq(A,B)/N
        freqOg = apriori(df1, min_support=0.001, use_colnames=True, low_memory=True)
        freqAnon = apriori(df2, min_support=0.001, use_colnames=True, low_memory=True)
        freqKOg = set({})
        freqKAnon = set({})
        #No single item allowed
        freqOg['length'] = freqOg['itemsets'].apply(lambda x: len(x))
        freqAnon['length'] = freqAnon['itemsets'].apply(lambda x: len(x))
        freqOg = freqOg[(freqOg['length']>=2)]
        freqAnon = freqAnon[(freqAnon['length']>=2)]
        #Top-K frequent itemsets of original dataset
        for itemset in freqOg['itemsets']:
            freqKOg.add(itemset)
            if len(freqKOg) == k:
                break

        #Top-K frequent itemsets of anonymized dataset
        for itemset in freqAnon['itemsets']:
            freqKAnon.add(itemset)
            if len(freqKAnon) == k:
                break

        #Amount of top-K frequent itemsets of the original dataset appeared
        #in the top-K frequent itemsets of the anonymized dataset
        intersect = 0
        for itemset in freqKOg:
            if itemset in freqKAnon:
                intersect+=1
        if len(freqKOg)<=0:
            tkd = 0
            print("It is empty")

        else:
            print(len(freqKOg))
            print(intersect)
            tkd = 1 - (intersect / len(freqKOg))
          
        return tkd
        
        
            
            
        
        

    
